from typing import Union

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from rest_framework import permissions
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework.request import Request

from .scopes import Scope


class ActionScopesRequired(permissions.BasePermission):
    """
    Look at the scopes required for the current action and check that they
    are present in the JWT
    """

    def get_required_scopes(self, view) -> Union[Scope, None]:
        if not hasattr(view, 'required_scopes'):
            raise ImproperlyConfigured("The View(Set) must have a `required_scopes` attribute")

        scopes_required = view.required_scopes.get(view.action)
        return scopes_required

    def has_permission(self, request: Request, view) -> bool:
        if settings.DEBUG and isinstance(request.accepted_renderer, BrowsableAPIRenderer):
            return True

        scopes_needed = self.get_required_scopes(view)
        # TODO: if no scopes are needed, what do???
        return request.jwt_payload.has_scopes(scopes_needed)



import re
from typing import Union

from django.db import models
from django.http import HttpRequest
from django.urls import Resolver404, get_resolver

from djangorestframework_camel_case.util import underscoreToCamel

RE_UNDERSCORE = re.compile(r"[a-z]_[a-z]")


def lookup_kwargs_to_filters(lookup_kwargs: dict, kwargs: dict) -> dict:
    """
    Using the lookup_kwargs map and the view kwargs, construct the queryset
    filter dict.
    """
    filters = {}
    for kwarg, field_name in lookup_kwargs.items():
        if kwarg not in kwargs:
            continue
        filters[field_name] = kwargs[kwarg]
    return filters


def get_viewset_for_path(path: str) -> 'rest_framework.viewsets.ViewSet':
    """
    Look up which viewset matches a path.
    """
    # NOTE: this doesn't support setting a different urlconf on the request
    resolver = get_resolver()
    try:
        resolver_match = resolver.resolve(path)
    except Resolver404 as exc:
        raise models.ObjectDoesNotExist("URL did not resolve") from exc
    callback, callback_args, callback_kwargs = resolver_match

    assert hasattr(callback, 'cls'), "Callback doesn't appear to be from a viewset"

    viewset = callback.cls(**callback.initkwargs)
    viewset.action_map = callback.actions
    viewset.request = HttpRequest()
    viewset.args = callback_args
    viewset.kwargs = callback_kwargs

    return viewset


def get_resource_for_path(path: str) -> models.Model:
    """
    Retrieve the API instance belonging to a (detail) path.
    """
    viewset = get_viewset_for_path(path)

    # See rest_framework.mixins.RetieveModelMixin.get_object()
    lookup_url_kwarg = viewset.lookup_url_kwarg or viewset.lookup_field
    filter_kwargs = {viewset.lookup_field: viewset.kwargs[lookup_url_kwarg]}

    return viewset.get_queryset().get(**filter_kwargs)


def underscore_to_camel(input_: Union[str, int]) -> str:
    """
    Convert a string from under_score to camelCase.
    """
    if not isinstance(input_, str):
        return input_

    return re.sub(RE_UNDERSCORE, underscoreToCamel, input_)



from rest_framework.exceptions import NotAcceptable
from rest_framework.renderers import BrowsableAPIRenderer

from .exceptions import PreconditionFailed

HEADER_ACCEPT = 'Accept-Crs'
HEADER_CONTENT = 'Content-Crs'

DEFAULT_CRS = 'EPSG:4326'


def extract_header(request, name: str) -> str:
    _header = name.replace('-', '_').upper()
    header = f'HTTP_{_header}'
    return request.META.get(header)


class GeoMixin:
    """
    GeoJSON viewset mixin.
    """
    @property
    def default_response_headers(self):
        headers = super().default_response_headers
        headers['Content-Crs'] = DEFAULT_CRS
        return headers

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        self.perform_crs_negotation(request)

    def perform_crs_negotation(self, request):
        # don't cripple the browsable API...
        if isinstance(request.accepted_renderer, BrowsableAPIRenderer):
            return

        # methods with request bodies need to have the CRS specified
        if request.method.lower() in ('post', 'put', 'patch'):
            content_crs = extract_header(request, HEADER_CONTENT)
            if content_crs is None:
                raise PreconditionFailed(
                    detail=F"'{HEADER_CONTENT}' header ontbreekt",
                )
            if content_crs != DEFAULT_CRS:
                raise NotAcceptable(
                    detail=f"CRS '{content_crs}' is niet ondersteund",
                )

        # client must indicate which CRS they want in the response
        requested_crs = extract_header(request, HEADER_ACCEPT)
        if requested_crs is None:
            raise PreconditionFailed(
                detail=F"'{HEADER_ACCEPT}' header ontbreekt",
            )

        if requested_crs != DEFAULT_CRS:
            raise NotAcceptable(
                detail=f"CRS '{requested_crs}' is niet ondersteund",
            )



import logging
import uuid
from collections import OrderedDict
from typing import List, Union

from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.urls import reverse

from rest_framework import exceptions

from .serializers import FoutSerializer, ValidatieFoutSerializer
from .utils import underscore_to_camel

logger = logging.getLogger(__name__)

ErrorSerializer = Union[FoutSerializer, ValidatieFoutSerializer]

STATUS_TO_TITLE = {}  # TODO


def _translate_exceptions(exc):
    # Taken from DRF default exc handler
    if isinstance(exc, Http404):
        exc = exceptions.NotFound()
    elif isinstance(exc, PermissionDenied):
        exc = exceptions.PermissionDenied()
    return exc


def get_validation_errors(validation_errors: dict):
    for field_name, error_list in validation_errors.items():

        # nested validation - recursively call the function
        if isinstance(error_list, dict):
            for err in get_validation_errors(error_list):
                err['name'] = f"{underscore_to_camel(field_name)}.{err['name']}"
                yield err
            continue

        if isinstance(error_list, exceptions.ErrorDetail):
            error_list = [error_list]

        for error in error_list:
            yield OrderedDict([
                # see https://tools.ietf.org/html/rfc7807#section-3.1
                # ('type', 'about:blank'),
                ('name', underscore_to_camel(field_name)),
                ('code', error.code),
                ('reason', str(error)),
            ])


class HandledException:

    def __init__(self, exc: exceptions.APIException, response, request=None):
        self.exc = exc
        assert 400 <= response.status_code < 600, "Unsupported status code"
        self.response = response
        self.request = request

        self._exc_id = str(uuid.uuid4())

    @property
    def _error_detail(self) -> str:
        if isinstance(self.exc, exceptions.ValidationError):
            # ErrorDetail from DRF is a str subclass
            data = getattr(self.response, 'data', {})
            return data.get('detail', '')
        # any other exception -> return the raw ErrorDetails object so we get
        # access to the code later
        return self.exc.detail

    @classmethod
    def as_serializer(cls, exc: exceptions.APIException, response, request=None) -> ErrorSerializer:
        """
        Return the appropriate serializer class instance.
        """
        exc = _translate_exceptions(exc)
        self = cls(exc, response, request)
        self.log()

        if isinstance(exc, exceptions.ValidationError):
            serializer_class = ValidatieFoutSerializer
        else:
            serializer_class = FoutSerializer

        return serializer_class(instance=self)

    def log(self):
        logger.exception("Exception %s ocurred", self._exc_id)

    @property
    def type(self) -> str:
        exc_detail_url = reverse(
            'zds_schema:error-detail',
            kwargs={'exception_class': self.exc.__class__.__name__}
        )
        if self.request is not None:
            exc_detail_url = self.request.build_absolute_uri(exc_detail_url)
        return exc_detail_url

    @property
    def code(self) -> str:
        if isinstance(self.exc, exceptions.ValidationError):
            return self.exc.default_code
        return self._error_detail.code if self._error_detail else ''

    @property
    def title(self) -> str:
        """
        Return the generic message for this type of exception.
        """
        default_title = getattr(self.exc, 'default_detail', str(self._error_detail))
        title = STATUS_TO_TITLE.get(self.response.status_code, default_title)
        return title

    @property
    def status(self) -> int:
        return self.response.status_code

    @property
    def detail(self) -> str:
        return str(self._error_detail)

    @property
    def instance(self) -> str:
        return f"urn:uuid:{self._exc_id}"

    @property
    def invalid_params(self) -> Union[None, List]:
        return [
            error for error in
            get_validation_errors(self.exc.detail)
        ]



"""
Utility module for Open API Specification 3.0.x.

This should get merged into gemma-zds-client, but some heavy refactoring is
needed for that.
"""
from typing import Union

import requests
import yaml
from drf_yasg import openapi

TYPE_MAP = {
    openapi.TYPE_OBJECT: dict,
    openapi.TYPE_STRING: str,
    openapi.TYPE_NUMBER: (float, int),
    openapi.TYPE_INTEGER: int,
    openapi.TYPE_BOOLEAN: bool,
    openapi.TYPE_ARRAY: list,
}


class SchemaFetcher:

    def __init__(self):
        self.cache = {}

    def fetch(self, url: str):
        """
        Fetch a YAML-based OAS 3.0.x schema.
        """
        if url in self.cache:
            return self.cache[url]

        response = requests.get(url)
        response.raise_for_status()

        spec = yaml.safe_load(response.content)
        spec_version = response.headers.get('X-OAS-Version', spec.get('openapi', spec.get('swagger', '')))
        if not spec_version.startswith('3.0'):
            raise ValueError("Unsupported spec version: {}".format(spec_version))

        self.cache[url] = spec
        return spec


def obj_has_shape(obj: Union[list, dict], schema: dict, resource: str) -> bool:
    """
    Compare an instance of an object with the expected shape from an OAS 3 schema.

    ..todo:: doesn't handle references and nested schema's yet.

    :param obj: the value retrieved from the endpoint, json-decoded to a dict or list
    :param schema: the OAS 3.0.x schema to test against, yaml-decoded to dict
    :param resource: the name of the resource to test the schape against
    """
    obj_schema = schema['components']['schemas'][resource]

    required = obj_schema.get('required', [])
    for prop, prop_schema in obj_schema['properties'].items():
        if prop in required and prop not in obj:
            # missing required prop -> can't match the schema
            return False

        # can't compare something that isn't there...
        if prop not in obj:
            continue

        value = obj[prop]
        expected_type = TYPE_MAP[prop_schema['type']]

        # type mismatch -> not what we're looking for
        if not isinstance(value, expected_type):
            return False

    return True


fetcher = SchemaFetcher()



import datetime

from django.db import transaction

import isodate
from djchoices import DjangoChoices
from rest_framework import fields, serializers

from .descriptors import GegevensGroepType


class DurationField(fields.DurationField):

    def to_internal_value(self, value):
        if isinstance(value, datetime.timedelta):
            return value
        try:
            parsed = isodate.parse_duration(str(value))
        except isodate.ISO8601Error:
            self.fail('invalid', format='P(n)Y(n)M(n)D')
        else:
            assert isinstance(parsed, datetime.timedelta)
            return parsed

    def to_representation(self, value):
        return isodate.duration_isoformat(value)


class FieldValidationErrorSerializer(serializers.Serializer):
    """
    Formaat van validatiefouten.
    """
    name = serializers.CharField(help_text="Naam van het veld met ongeldige gegevens")
    code = serializers.CharField(help_text="Systeemcode die het type fout aangeeft")
    reason = serializers.CharField(help_text="Uitleg wat er precies fout is met de gegevens")


class FoutSerializer(serializers.Serializer):
    """
    Formaat van HTTP 4xx en 5xx fouten.
    """
    type = serializers.CharField(
        help_text="URI referentie naar het type fout, bedoeld voor developers",
        required=False, allow_blank=True
    )
    # not according to DSO, but possible for programmatic checking
    code = serializers.CharField(help_text="Systeemcode die het type fout aangeeft")
    title = serializers.CharField(help_text="Generieke titel voor het type fout")
    status = serializers.IntegerField(help_text="De HTTP status code")
    detail = serializers.CharField(help_text="Extra informatie bij de fout, indien beschikbaar")
    instance = serializers.CharField(
        help_text="URI met referentie naar dit specifiek voorkomen van de fout. Deze kan "
                  "gebruikt worden in combinatie met server logs, bijvoorbeeld."
    )


class ValidatieFoutSerializer(FoutSerializer):
    pass


# can't declare stuff with dashes and DSO prescribes dashed key...
ValidatieFoutSerializer._declared_fields['invalid-params'] = \
    FieldValidationErrorSerializer(source='invalid_params', many=True)


def add_choice_values_help_text(choices: DjangoChoices) -> str:
    displays = "\n".join([
        f"* `{value}` - {display}"
        for value, display in choices.choices
    ])
    return f"De mapping van waarden naar weergave is als volgt:\n\n{displays}"


class GegevensGroepSerializerMetaclass(serializers.SerializerMetaclass):

    def __new__(cls, name, bases, attrs):
        Meta = attrs.get('Meta')
        if Meta:
            assert hasattr(Meta, 'model'), "The 'model' class must be defined on the Meta."
            assert hasattr(Meta, 'gegevensgroep'), "The 'gegevensgroep' name must be defined on the Meta."

            gegevensgroep = getattr(Meta.model, Meta.gegevensgroep)
            Meta.fields = []

            extra_kwargs = {}

            for field_name, model_field in gegevensgroep.mapping.items():
                Meta.fields.append(field_name)

                # the field is always required and may not be empty in any form
                default_extra_kwargs = {
                    'source': model_field.name,
                    'required': field_name not in gegevensgroep.optional,
                    'allow_null': False,
                    'allow_blank': field_name in gegevensgroep.optional,
                }

                internal_type = model_field.get_internal_type()
                if internal_type not in ['CharField', 'TextField']:
                    del default_extra_kwargs['allow_blank']
                if internal_type == 'BooleanField':
                    del default_extra_kwargs['allow_null']

                extra_kwargs[field_name] = default_extra_kwargs

                declared_extra_kwargs = getattr(Meta, 'extra_kwargs', {}).get(field_name)
                if declared_extra_kwargs:
                    extra_kwargs[field_name].update(declared_extra_kwargs)

            Meta.extra_kwargs = extra_kwargs

        return super().__new__(cls, name, bases, attrs)


class GegevensGroepSerializer(serializers.ModelSerializer, metaclass=GegevensGroepSerializerMetaclass):
    """
    Generate a serializer out of a GegevensGroepType.

    Usage::

    >>> class VerlengingSerializer(GegevensGroepSerializer):
    ...     class Meta:
    ...         model = Zaak
    ...         gegevensgroep = 'verlenging'
    >>>

    Where ``Zaak.verlenging`` is a :class:``GegevensGroepType``.
    """

    def to_representation(self, instance) -> dict:
        """
        Output the result of accessing the descriptor.
        """
        return instance

    def to_internal_value(self, data):
        """
        Pass through the original keys instead of reverse-mapping the source attrs.
        """
        value = super().to_internal_value(data)
        ret = {}
        for name, field in self.fields.items():
            if field not in self._writable_fields:
                continue
            if field.source not in value:
                continue
            ret[name] = value[field.source]
        return ret


class NestedGegevensGroepMixin:
    """
    Set gegevensgroepdata from validated nested data.

    Usage: include the mixin on the ModelSerializer that has gegevensgroepen.
    """

    def _is_gegevensgroep(self, name: str):
        attr = getattr(self.Meta.model, name)
        return isinstance(attr, GegevensGroepType)

    @transaction.atomic
    def create(self, validated_data):
        """
        Handle nested writes.
        """
        gegevensgroepen = {}

        for name in list(validated_data.keys()):
            if not self._is_gegevensgroep(name):
                continue

            gegevensgroepen[name] = validated_data.pop(name)

        # perform the default create
        obj = super().create(validated_data)

        for name, gegevensgroepdata in gegevensgroepen.items():
            setattr(obj, name, gegevensgroepdata)

        obj.save()

        return obj

    def update(self, instance, validated_data):
        """
        Handle nested writes.
        """
        for name in list(validated_data.keys()):
            if not self._is_gegevensgroep(name):
                continue
            setattr(instance, name, validated_data.pop(name))
        return super().update(instance, validated_data)



from typing import Dict

from django.db import models


class GegevensGroepType:
    """
    Structure a body of data from flat model fields.

    All keys are always truthy for a value.

    :param mapping: dict, mapping simple keys to model fields
    :param optional: iterable of fields that are NOT required.
    """
    name = None
    model = None

    def __init__(self, mapping: Dict[str, models.Field], optional: tuple=None):
        self.mapping = mapping
        self.optional = optional or ()

        all_fields_known = set(self.optional).issubset(set(mapping.keys()))
        assert all_fields_known, "The fields in 'optional' must be a subset of the mapping keys"

        # check if it's optional or not
        self.required = any(field.blank is False for field in self.mapping.values())

    def __repr__(self):
        fields = ", ".join([
            field if field not in self.optional else f"{field} (optional)"
            for field in self.mapping.keys()
        ])
        return "<GegevensGroepType: fields=%r required=%r>" % (fields, self.required)

    def __get__(self, obj, type=None):
        if obj is None:  # accessed through the owner, i.e. the model -> introspection
            return self

        return {
            key: getattr(obj, field.name)
            for key, field in self.mapping.items()
        }

    def __set__(self, obj, value: dict):
        # value can be empty, if that's the case, empty all model fields
        if not value:
            if self.required:
                raise ValueError("A non-empty value is required")

            for field in self.mapping.values():
                empty_value = None if field.null else ''
                default_value = field.default if field.default != models.NOT_PROVIDED else empty_value
                setattr(obj, field.name, default_value)
            return

        # map the values
        for key, field in self.mapping.items():
            setattr(obj, field.name, value[key])

            if key not in self.optional:
                assert getattr(obj, field.name, None), f"Empty '{key}' not allowed"



from typing import List

OPERATOR_OR = 'OR'
OPERATOR_AND = 'AND'


SCOPE_REGISTRY = set()


class Scope:

    def __init__(self, label: str, description: str=None):
        self.label = label
        self.description = description

        # combined scopes
        self.children = []
        self.operator = None

        # add to registry
        SCOPE_REGISTRY.add(self)

    def __repr__(self) -> str:
        if self.children:
            return "<%s: %r" % (self.operator, self.children)

        cls_name = self.__class__.__name__
        return "<%s: label=%r>" % (cls_name, self.label)

    def __str__(self):
        return f"({self.label})" if self.children else self.label

    def __or__(self, other):
        new = type(self)(label=f"{self.label} | {other.label}")
        new.children = [self, other]
        new.operator = OPERATOR_OR
        return new

    def is_contained_in(self, scope_set: List[str]) -> bool:
        """
        Test if the flat ``scope_set`` encapsulate this scope.
        """
        if not self.children:
            return self.label in scope_set

        children_contained = (
            child.is_contained_in(scope_set)
            for child in self.children
        )

        if self.operator == OPERATOR_OR:
            return any(children_contained)
        elif self.operator == OPERATOR_AND:
            return all(children_contained)
        else:
            raise ValueError(f"Unkonwn operator '{self.operator}'")



import logging
import os

from django.apps import apps
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.template.loader import render_to_string
from django.urls import reverse

from drf_yasg import openapi
from drf_yasg.app_settings import swagger_settings
from drf_yasg.management.commands import generate_swagger
from rest_framework.settings import api_settings

from ...schema import OpenAPISchemaGenerator
from ...version import get_major_version


class Table:
    def __init__(self, resource: str):
        self.resource = resource
        self.rows = []


class Row:
    def __init__(self,
                 label: str, description: str, type: str, required: bool,
                 create: bool, read: bool, update: bool, delete: bool):
        self.label = label
        self.description = description
        self.type = type
        self.required = required
        self.create = create
        self.read = read
        self.update = update
        self.delete = delete


class Command(generate_swagger.Command):
    """
    Patches to the provided command to modify the schema for ZDS needs.
    """

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument('--to-markdown-table', action='store_true')

    def get_mock_request(self, *args, **kwargs):
        request = super().get_mock_request(*args, **kwargs)
        request.version = api_settings.DEFAULT_VERSION
        return request

    def write_schema(self, schema, stream, format):
        del schema.host
        del schema.schemes
        super().write_schema(schema, stream, format)

    # need to overwrite the generator class...
    def handle(self, output_file, overwrite, format, api_url, mock, user, private, *args, **options):
        # disable logs of WARNING and below
        logging.disable(logging.WARNING)

        info = getattr(swagger_settings, 'DEFAULT_INFO', None)
        if not isinstance(info, openapi.Info):
            raise ImproperlyConfigured(
                'settings.SWAGGER_SETTINGS["DEFAULT_INFO"] should be an '
                'import string pointing to an openapi.Info object'
            )

        if not format:
            if os.path.splitext(output_file)[1] in ('.yml', '.yaml'):
                format = 'yaml'
        format = format or 'json'

        api_root = reverse('api-root', kwargs={'version': get_major_version()})
        api_url = (
            api_url
            or swagger_settings.DEFAULT_API_URL  # noqa
            or f"http://example.com{api_root}"  # noqa
        )

        user = User.objects.get(username=user) if user else None
        mock = mock or private or (user is not None)
        if mock and not api_url:
            raise ImproperlyConfigured(
                '--mock-request requires an API url; either provide '
                'the --url argument or set the DEFAULT_API_URL setting'
            )

        request = self.get_mock_request(api_url, format, user) if mock else None

        generator = OpenAPISchemaGenerator(
            info=info,
            url=api_url
        )
        schema = generator.get_schema(request=request, public=not private)

        if output_file == '-':
            self.write_schema(schema, self.stdout, format)
        else:
            with open(output_file, 'w', encoding='utf8') as stream:
                if options['to_markdown_table']:
                    self.to_markdown_table(schema, stream)
                else:
                    self.write_schema(schema, stream, format)

    def to_markdown_table(self, schema, stream):
        template = 'zds_schema/api_schema_to_markdown_table.md'
        tables = []

        whitelist = [model._meta.object_name for model in apps.get_models()]

        for resource, definition in schema.definitions.items():
            if resource not in whitelist:
                continue

            if not hasattr(definition, 'properties'):
                continue

            table = Table(resource)
            for field, _schema in definition.properties.items():
                if isinstance(_schema, openapi.SchemaRef):
                    continue

                readonly = getattr(_schema, 'readOnly', False)
                table.rows.append(Row(
                    label=field,
                    description=getattr(_schema, 'description', ''),
                    type=_schema.type,
                    required=field in definition.required,
                    create=not readonly,
                    read=True,
                    update=not readonly,
                    delete=not readonly,
                ))
            tables.append(table)

        markdown = render_to_string(template, context={'tables': tables})
        stream.write(markdown)



from django.utils.translation import ugettext_lazy as _

from djchoices import ChoiceItem, DjangoChoices

BSN_LENGTH = 9
RSIN_LENGTH = 9

VERSION_HEADER = 'API-version'


class VertrouwelijkheidsAanduiding(DjangoChoices):
    openbaar = ChoiceItem('openbaar', 'OPENBAAR')
    beperkt_openbaar = ChoiceItem('beperkt openbaar', 'BEPERKT OPENBAAR')
    intern = ChoiceItem('intern', 'INTERN')
    zaakvertrouwelijk = ChoiceItem('zaakvertrouwelijk', 'ZAAKVERTROUWELIJK')
    vertrouwelijk = ChoiceItem('vertrouwelijk', 'VERTROUWELIJK')
    confidentieel = ChoiceItem('confidentieel', 'CONFIDENTIEEL')
    geheim = ChoiceItem('geheim', 'GEHEIM')
    zeer_geheim = ChoiceItem('zeer geheim', 'ZEER GEHEIM')


class RolOmschrijving(DjangoChoices):
    adviseur = ChoiceItem(
        'Adviseur', 'Adviseur',
        description='Kennis in dienst stellen van de behandeling van (een deel van) een zaak.'
    )
    behandelaar = ChoiceItem(
        'Behandelaar', 'Behandelaar',
        description='De vakinhoudelijke behandeling doen van (een deel van) een zaak.'
    )
    belanghebbende = ChoiceItem(
        'Belanghebbende', 'Belanghebbende',
        description='Vanuit eigen en objectief belang rechtstreeks betrokken '
                    'zijn bij de behandeling en/of de uitkomst van een zaak.'
    )
    beslisser = ChoiceItem(
        'Beslisser', 'Beslisser',
        description='Nemen van besluiten die voor de uitkomst van een zaak noodzakelijk zijn.'
    )
    initiator = ChoiceItem(
        'Initiator', 'Initiator',
        description='Aanleiding geven tot de start van een zaak ..'
    )
    klantcontacter = ChoiceItem(
        'Klantcontacter', 'Klantcontacter',
        description='Het eerste aanspreekpunt zijn voor vragen van burgers en bedrijven ..'
    )
    zaakcoordinator = ChoiceItem(
        'Zaakcoördinator', 'Zaakcoördinator',
        description='Er voor zorg dragen dat de behandeling van de zaak in samenhang '
                    'uitgevoerd wordt conform de daarover gemaakte afspraken.'
    )
    medeinitiator = ChoiceItem(
        'Mede-initiator', 'Mede-initiator',
        description=''
    )


class RolTypes(DjangoChoices):
    natuurlijk_persoon = ChoiceItem('Natuurlijk persoon', "Natuurlijk persoon")
    niet_natuurlijk_persoon = ChoiceItem('Niet-natuurlijk persoon', "Niet-natuurlijk persoon")
    vestiging = ChoiceItem('Vestiging', "Vestiging")
    organisatorische_eenheid = ChoiceItem('Organisatorische eenheid', "Organisatorische eenheid")
    medewerker = ChoiceItem('Medewerker', "Medewerker")


class ObjectTypes(DjangoChoices):
    besluit = ChoiceItem('besluit', _("Besluit"))
    zaak = ChoiceItem('zaak', _("Zaak"))


class Archiefnominatie(DjangoChoices):
    blijvend_bewaren = ChoiceItem(
        'blijvend_bewaren',
        _("Het zaakdossier moet bewaard blijven en op de Archiefactiedatum overgedragen worden naar een "
          "archiefbewaarplaats.")
    )
    vernietigen = ChoiceItem(
        'vernietigen',
        _("Het zaakdossier moet op of na de Archiefactiedatum vernietigd worden.")
    )


class Archiefstatus(DjangoChoices):
    nog_te_archiveren = ChoiceItem(
        'nog_te_archiveren',
        _("De zaak cq. het zaakdossier is nog niet als geheel gearchiveerd.")
    )
    gearchiveerd = ChoiceItem(
        'gearchiveerd',
        _("De zaak cq. het zaakdossier is als geheel niet-wijzigbaar bewaarbaar gemaakt.")
    )
    gearchiveerd_procestermijn_onbekend = ChoiceItem(
        'gearchiveerd_procestermijn_onbekend',
        _("De zaak cq. het zaakdossier is als geheel niet-wijzigbaar bewaarbaar gemaakt maar de vernietigingsdatum "
          "kan nog niet bepaald worden.")
    )
    # After deliberation this element was removed because "vernietigd" means
    # it's really gone and the status wouldn't make sense:
    #
    # vernietigd = ChoiceItem('vernietigd',
    #     _("De zaak cq. het zaakdossier is vernietigd.")
    # )
    overgedragen = ChoiceItem(
        'overgedragen',
        _("De zaak cq. het zaakdossier is overgebracht naar een archiefbewaarplaats.")
    )


class BrondatumArchiefprocedureAfleidingswijze(DjangoChoices):
    afgehandeld = ChoiceItem(
        'afgehandeld', _("Afgehandeld"),
        description=_("De termijn start op de datum waarop de zaak is "
                      "afgehandeld (ZAAK.Einddatum in het RGBZ).")
    )
    ander_datumkenmerk = ChoiceItem(
        'ander_datumkenmerk', _("Ander datumkenmerk"),
        description=_("De termijn start op de datum die is vastgelegd in een "
                      "ander datumveld dan de datumvelden waarop de overige "
                      "waarden (van deze attribuutsoort) betrekking hebben. "
                      "`Objecttype`, `Registratie` en `Datumkenmerk` zijn niet "
                      "leeg.")
    )
    eigenschap = ChoiceItem(
        'eigenschap', _("Eigenschap"),
        description=_("De termijn start op de datum die vermeld is in een "
                      "zaaktype-specifieke eigenschap (zijnde een `datumveld`). "
                      "`ResultaatType.ZaakType` heeft een `Eigenschap`; "
                      "`Objecttype`, en `Datumkenmerk` zijn niet leeg.")
    )
    gerelateerde_zaak = ChoiceItem(
        'gerelateerde_zaak', _("Gerelateerde zaak"),
        description=_("De termijn start op de datum waarop de gerelateerde "
                      "zaak is afgehandeld (`ZAAK.Einddatum` of "
                      "`ZAAK.Gerelateerde_zaak.Einddatum` in het RGBZ). "
                      "`ResultaatType.ZaakType` heeft gerelateerd `ZaakType`")
    )
    hoofdzaak = ChoiceItem(
        'hoofdzaak', _("Hoofzaak"),
        description=_("De termijn start op de datum waarop de gerelateerde "
                      "zaak is afgehandeld, waarvan de zaak een deelzaak is "
                      "(`ZAAK.Einddatum` van de hoofdzaak in het RGBZ). "
                      "ResultaatType.ZaakType is deelzaaktype van ZaakType.")
    )
    ingangsdatum_besluit = ChoiceItem(
        'ingangsdatum_besluit', _("Ingangsdatum besluit"),
        description=_("De termijn start op de datum waarop het besluit van "
                      "kracht wordt (`BESLUIT.Ingangsdatum` in het RGBZ).	"
                      "ResultaatType.ZaakType heeft relevant BesluitType")
    )
    termijn = ChoiceItem(
        'termijn', _("Termijn"),
        description=_("De termijn start een vast aantal jaren na de datum "
                      "waarop de zaak is afgehandeld (`ZAAK.Einddatum` in het "
                      "RGBZ).")
    )
    vervaldatum_besluit = ChoiceItem(
        'vervaldatum_besluit', _("Vervaldatum besluit"),
        description=_("De termijn start op de dag na de datum waarop het "
                      "besluit vervalt (`BESLUIT.Vervaldatum` in het RGBZ). "
                      "ResultaatType.ZaakType heeft relevant BesluitType")
    )
    zaakobject = ChoiceItem(
        'zaakobject', _("Zaakobject"),
        description=_("De termijn start op de einddatum geldigheid van het "
                      "zaakobject waarop de zaak betrekking heeft "
                      "(bijvoorbeeld de overlijdendatum van een Persoon). "
                      "M.b.v. de attribuutsoort `Objecttype` wordt vastgelegd "
                      "om welke zaakobjecttype het gaat; m.b.v. de "
                      "attribuutsoort `Datumkenmerk` wordt vastgelegd welke "
                      "datum-attribuutsoort van het zaakobjecttype het betreft.")
    )



import inspect

from django.conf import settings
from django.db import models
from django.urls import reverse as _reverse, reverse_lazy as _reverse_lazy


def _magic_args(args, kwargs):
    """
    Do some trivial introspection to translate objects/models into common
    used urls.
    """
    if args and isinstance(args[0], models.Model):
        url_name = f"{args[0]._meta.model_name}-detail"
        kwargs['kwargs'].setdefault('uuid', args[0].uuid)
        args = (url_name,) + args[1:]
    elif args and inspect.isclass(args[0]) and issubclass(args[0], models.Model):
        url_name = f"{args[0]._meta.model_name}-list"
        args = (url_name,) + args[1:]
    return args, kwargs


def reverse(*args, **kwargs):
    kwargs.setdefault('kwargs', {})
    kwargs['kwargs']['version'] = settings.REST_FRAMEWORK['DEFAULT_VERSION']
    args, kwargs = _magic_args(args, kwargs)
    return _reverse(*args, **kwargs)


def reverse_lazy(*args, **kwargs):
    kwargs.setdefault('kwargs', {})
    kwargs['kwargs']['version'] = settings.REST_FRAMEWORK['DEFAULT_VERSION']
    args, kwargs = _magic_args(args, kwargs)
    return _reverse_lazy(*args, **kwargs)



from urllib.parse import urlencode, urlparse

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.forms.widgets import URLInput
from django.http import QueryDict
from django.utils.translation import ugettext_lazy as _

from django_filters import fields, filters
from django_filters.constants import EMPTY_VALUES
from django_filters.rest_framework import DjangoFilterBackend
from djangorestframework_camel_case.parser import CamelCaseJSONParser
from djangorestframework_camel_case.render import CamelCaseJSONRenderer
from djangorestframework_camel_case.util import underscoreize
from rest_framework.request import Request
from rest_framework.views import APIView

from .search import is_search_view
from .utils import get_resource_for_path
from .validators import validate_rsin


class Backend(DjangoFilterBackend):

    # Taken from drf_yasg.inspectors.field.CamelCaseJSONFilter
    def _is_camel_case(self, view):
        return (
            any(issubclass(parser, CamelCaseJSONParser) for parser in view.parser_classes) or
            any(issubclass(renderer, CamelCaseJSONRenderer) for renderer in view.renderer_classes)
        )

    def _transform_query_params(self, view, query_params: QueryDict) -> QueryDict:
        if not self._is_camel_case(view):
            return query_params

        # data can be a regular dict if it's coming from a serializer
        if hasattr(query_params, 'lists'):
            data = dict(query_params.lists())
        else:
            data = query_params

        transformed = underscoreize(data)

        return QueryDict(urlencode(transformed, doseq=True))

    def get_filterset_kwargs(self, request: Request, queryset: models.QuerySet, view: APIView):
        """
        Get the initialization parameters for the filterset.

        * filter on request.data if request.query_params is empty
        * do the camelCase transformation of filter parameters
        """
        kwargs = super().get_filterset_kwargs(request, queryset, view)
        filter_parameters = request.query_params if not is_search_view(view) else request.data
        query_params = self._transform_query_params(view, filter_parameters)
        kwargs['data'] = query_params
        return kwargs


class URLModelChoiceField(fields.ModelChoiceField):
    widget = URLInput

    def url_to_pk(self, url: str):
        parsed = urlparse(url)
        path = parsed.path
        if settings.FORCE_SCRIPT_NAME and path.startswith(settings.FORCE_SCRIPT_NAME):
            path = path[len(settings.FORCE_SCRIPT_NAME):]
        instance = get_resource_for_path(path)
        model = self.queryset.model
        if not isinstance(instance, model):
            raise ValidationError(_("Invalid resource type supplied, expected %r") % model, code='invalid-type')
        return instance.pk

    def to_python(self, value: str):
        # TODO: validate that it's proper URL input
        if value:
            try:
                value = self.url_to_pk(value)
            except models.ObjectDoesNotExist:
                raise ValidationError(_("Invalid resource URL supplied"), code='invalid')
        return super().to_python(value)


class URLModelChoiceFilter(filters.ModelChoiceFilter):
    field_class = URLModelChoiceField


class RSINFilter(filters.CharFilter):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('validators', [validate_rsin])
        super().__init__(*args, **kwargs)


class WildcardFilter(filters.CharFilter):
    """
    Filters the queryset based on a string and optionally allows wildcards in
    the query parameter.
    """
    wildcard = '*'

    def __init__(self, *args, **kwargs):
        kwargs['lookup_expr'] = 'iregex'
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        value = r'^{}$'.format(value.replace(self.wildcard, '.*'))

        return super().filter(qs, value)



from django.urls import path

from .views import ErrorDetailView, ScopesView

app_name = 'zds_schema'

urlpatterns = [
    path('fouten/<exception_class>/', ErrorDetailView.as_view(), name='error-detail'),
    path('scopes/', ScopesView.as_view(), name='scopes'),
]



from django.apps import AppConfig
from django.db import models

from drf_yasg import openapi
from drf_yasg.inspectors.field import (
    basic_type_info, model_field_to_basic_type, serializer_field_to_basic_type
)
from rest_framework import serializers

from . import fields
from .serializers import DurationField


class ZDSSchemaConfig(AppConfig):
    name = 'zds_schema'

    def ready(self):
        patch_duration_type()
        register_serializer_field()


def patch_duration_type():

    def _patch(basic_types, _field_cls):
        for index, (field_cls, basic_type) in enumerate(basic_types):
            if field_cls is _field_cls:
                basic_types[index] = (_field_cls, (openapi.TYPE_STRING, None))
                break

    _patch(model_field_to_basic_type, models.DurationField)
    _patch(basic_type_info, models.DurationField)
    _patch(serializer_field_to_basic_type, serializers.DurationField)
    _patch(basic_type_info, serializers.DurationField)


def register_serializer_field():
    mapping = serializers.ModelSerializer.serializer_field_mapping
    mapping[models.fields.DurationField] = DurationField
    mapping[fields.DaysDurationField] = DurationField



from copy import deepcopy

from django.core.validators import URLValidator
from django.db import models

from django_filters.rest_framework import filterset

from .fields import RSINField
from .filters import RSINFilter, URLModelChoiceFilter

FILTER_FOR_DBFIELD_DEFAULTS = deepcopy(filterset.FILTER_FOR_DBFIELD_DEFAULTS)
FILTER_FOR_DBFIELD_DEFAULTS[models.ForeignKey]['filter_class'] = URLModelChoiceFilter

# register custom field(s)
FILTER_FOR_DBFIELD_DEFAULTS[RSINField] = {'filter_class': RSINFilter}
FILTER_FOR_DBFIELD_DEFAULTS[models.URLField]['extra'] = lambda f: {'validators': [URLValidator()]}


class FilterSet(filterset.FilterSet):
    """
    Allow foreign key fields to be filtered on resource URL.
    """
    FILTER_DEFAULTS = FILTER_FOR_DBFIELD_DEFAULTS
